@@ -11,8 +11,8 @@
 
 #include <njs_auto_config.h>
 
-#define NJS_VERSION                 "0.8.2"
-#define NJS_VERSION_NUMBER          0x000802
+#define NJS_VERSION                 "0.8.7"
+#define NJS_VERSION_NUMBER          0x000807
 
 
 #include <string.h>
@@ -46,12 +46,6 @@ typedef struct {
     uint64_t                        filler[2];
 } njs_opaque_value_t;
 
-typedef enum {
-    NJS_LOG_LEVEL_ERROR = 4,
-    NJS_LOG_LEVEL_WARN = 5,
-    NJS_LOG_LEVEL_INFO = 7,
-} njs_log_level_t;
-
 /* sizeof(njs_value_t) is 16 bytes. */
 #define njs_argument(args, n)                                                 \
     (njs_value_t *) ((u_char *) args + (n) * 16)
@@ -72,19 +66,14 @@ extern const njs_value_t            njs_value_undefined;
     ((n < nargs) ? njs_argument(args, n)                                      \
                  : (njs_value_assign(lvalue, &njs_value_undefined), lvalue))
 
-#define njs_vm_log(vm, fmt, ...)                                              \
-    njs_vm_logger(vm, NJS_LOG_LEVEL_INFO, fmt, ##__VA_ARGS__)
-#define njs_vm_warn(vm, fmt, ...)                                             \
-    njs_vm_logger(vm, NJS_LOG_LEVEL_WARN, fmt, ##__VA_ARGS__)
-#define njs_vm_err(vm, fmt, ...)                                              \
-    njs_vm_logger(vm, NJS_LOG_LEVEL_ERROR, fmt, ##__VA_ARGS__)
-
 #define njs_vm_error(vm, fmt, ...)                                            \
     njs_vm_error2(vm, 0, fmt, ##__VA_ARGS__)
 #define njs_vm_internal_error(vm, fmt, ...)                                   \
     njs_vm_error2(vm, 2, fmt, ##__VA_ARGS__)
 #define njs_vm_range_error(vm, fmt, ...)                                      \
     njs_vm_error2(vm, 3, fmt, ##__VA_ARGS__)
+#define njs_vm_ref_error(vm, fmt, ...)                                        \
+    njs_vm_error2(vm, 4, fmt, ##__VA_ARGS__)
 #define njs_vm_syntax_error(vm, fmt, ...)                                     \
     njs_vm_error2(vm, 5, fmt, ##__VA_ARGS__)
 #define njs_vm_type_error(vm, fmt, ...)                                       \
@@ -141,6 +130,18 @@ typedef enum {
 } njs_wellknown_symbol_t;
 
 
+typedef enum {
+#define njs_object_enum_kind(flags) (flags & 7)
+    NJS_ENUM_KEYS = 1,
+    NJS_ENUM_VALUES = 2,
+    NJS_ENUM_BOTH = 4,
+#define njs_object_enum(flags) (flags & (NJS_ENUM_STRING | NJS_ENUM_SYMBOL))
+    NJS_ENUM_STRING = 8,
+    NJS_ENUM_SYMBOL = 16,
+    NJS_ENUM_ENUMERABLE_ONLY = 32,
+} njs_object_enum_t;
+
+
 typedef enum {
     /*
      * Extern property type.
@@ -205,46 +206,13 @@ struct njs_external_s {
 };
 
 
-/*
- * NJS and event loops.
- *
- * njs_vm_ops_t callbacks are used to interact with the event loop environment.
- *
- * Functions get an external object as the first argument. The external
- * object is provided as the third argument to njs_vm_clone().
- *
- * The callbacks are expected to return to the VM the unique id of an
- * underlying event.  This id will be passed as the second argument to
- * njs_event_destructor() at the moment the VM wants to destroy it.
- *
- * When an underlying events fires njs_vm_post_event() should be invoked with
- * the value provided as vm_event.
- *
- * The events posted by njs_vm_post_event() are processed as soon as
- * njs_vm_run() is invoked. njs_vm_run() returns NJS_AGAIN until pending events
- * are present.
- */
-
-typedef void *                      njs_vm_event_t;
-typedef void *                      njs_host_event_t;
 typedef void *                      njs_external_ptr_t;
 
-typedef njs_host_event_t (*njs_set_timer_t)(njs_external_ptr_t external,
-    uint64_t delay, njs_vm_event_t vm_event);
-typedef void (*njs_event_destructor_t)(njs_external_ptr_t external,
-    njs_host_event_t event);
 typedef njs_mod_t *(*njs_module_loader_t)(njs_vm_t *vm,
     njs_external_ptr_t external, njs_str_t *name);
-typedef void (*njs_logger_t)(njs_vm_t *vm, njs_external_ptr_t external,
-    njs_log_level_t level, const u_char *start, size_t length);
-
-
-typedef struct {
-    njs_set_timer_t                 set_timer;
-    njs_event_destructor_t          clear_timer;
-    njs_module_loader_t             module_loader;
-    njs_logger_t                    logger;
-} njs_vm_ops_t;
+typedef void (*njs_rejection_tracker_t)(njs_vm_t *vm,
+    njs_external_ptr_t external, njs_bool_t is_handled, njs_value_t *promise,
+    njs_value_t *reason);
 
 
 typedef struct {
@@ -265,7 +233,6 @@ typedef struct {
 typedef struct {
     njs_external_ptr_t              external;
     njs_vm_shared_t                 *shared;
-    njs_vm_ops_t                    *ops;
     njs_vm_meta_t                   *metas;
     njs_module_t                    **addons;
     njs_str_t                       file;
@@ -275,11 +242,6 @@ typedef struct {
 
     njs_uint_t                      max_stack_size;
 
-    njs_log_level_t                 log_level;
-
-#define NJS_VM_OPT_UNHANDLED_REJECTION_IGNORE   0
-#define NJS_VM_OPT_UNHANDLED_REJECTION_THROW    1
-
 /*
  * interactive  - enables "interactive" mode.
  *  (REPL). Allows starting parent VM without cloning.
@@ -292,9 +254,6 @@ typedef struct {
  *   - Function constructors.
  * module        - ES6 "module" mode. Script mode is default.
  * ast           - print AST.
- * unhandled_rejection IGNORE | THROW - tracks unhandled promise rejections:
- *   - throwing inside a Promise without a catch block.
- *   - throwing inside in a finally or catch block.
  */
     uint8_t                         interactive;     /* 1 bit */
     uint8_t                         trailer;         /* 1 bit */
@@ -312,7 +271,6 @@ typedef struct {
 #ifdef NJS_DEBUG_GENERATOR
     uint8_t                         generator_debug; /* 1 bit */
 #endif
-    uint8_t                         unhandled_rejection;
 } njs_vm_opt_t;
 
 
@@ -338,34 +296,27 @@ NJS_EXPORT njs_vm_t *njs_vm_create(njs_vm_opt_t *options);
 NJS_EXPORT void njs_vm_destroy(njs_vm_t *vm);
 
 NJS_EXPORT njs_int_t njs_vm_compile(njs_vm_t *vm, u_char **start, u_char *end);
+NJS_EXPORT void njs_vm_set_module_loader(njs_vm_t *vm,
+    njs_module_loader_t module_loader, void *opaque);
 NJS_EXPORT njs_mod_t *njs_vm_add_module(njs_vm_t *vm, njs_str_t *name,
     njs_value_t *value);
 NJS_EXPORT njs_mod_t *njs_vm_compile_module(njs_vm_t *vm, njs_str_t *name,
     u_char **start, u_char *end);
 NJS_EXPORT njs_vm_t *njs_vm_clone(njs_vm_t *vm, njs_external_ptr_t external);
 
-NJS_EXPORT njs_vm_event_t njs_vm_add_event(njs_vm_t *vm,
-    njs_function_t *function, njs_uint_t once, njs_host_event_t host_ev,
-    njs_event_destructor_t destructor);
-NJS_EXPORT void njs_vm_del_event(njs_vm_t *vm, njs_vm_event_t vm_event);
-NJS_EXPORT njs_int_t njs_vm_post_event(njs_vm_t *vm, njs_vm_event_t vm_event,
+NJS_EXPORT njs_int_t njs_vm_enqueue_job(njs_vm_t *vm, njs_function_t *function,
     const njs_value_t *args, njs_uint_t nargs);
-
 /*
- * Returns 1 if async events are present.
+ * Executes a single pending job.
+ *  1 successful run.
+ *  NJS_OK pending job was not found.
+ *  NJS_ERROR some exception or internal error happens.
  */
-NJS_EXPORT njs_int_t njs_vm_waiting(njs_vm_t *vm);
+NJS_EXPORT njs_int_t njs_vm_execute_pending_job(njs_vm_t *vm);
+NJS_EXPORT njs_int_t njs_vm_pending(njs_vm_t *vm);
 
-/*
- * Returns 1 if posted events are ready to be executed.
- */
-NJS_EXPORT njs_int_t njs_vm_posted(njs_vm_t *vm);
-
-#define njs_vm_pending(vm)  (njs_vm_waiting(vm) || njs_vm_posted(vm))
-
-NJS_EXPORT njs_int_t njs_vm_unhandled_rejection(njs_vm_t *vm);
-
-NJS_EXPORT void *njs_vm_completions(njs_vm_t *vm, njs_str_t *expression);
+NJS_EXPORT void njs_vm_set_rejection_tracker(njs_vm_t *vm,
+    njs_rejection_tracker_t rejection_tracker, void *opaque);
 
 /*
  * Runs the specified function with provided arguments.
@@ -379,16 +330,6 @@ NJS_EXPORT njs_int_t njs_vm_call(njs_vm_t *vm, njs_function_t *function,
 NJS_EXPORT njs_int_t njs_vm_invoke(njs_vm_t *vm, njs_function_t *function,
     const njs_value_t *args, njs_uint_t nargs, njs_value_t *retval);
 
-/*
- * Runs posted events.
- *  NJS_OK successfully processed all posted events, no more events.
- *  NJS_AGAIN successfully processed all events, some posted events are
- *    still pending.
- *  NJS_ERROR some exception or internal error happens.
- *    njs_vm_exception_get(vm) can be used to get the exception value.
- */
-NJS_EXPORT njs_int_t njs_vm_run(njs_vm_t *vm);
-
 /*
  * Runs the global code.
  *   NJS_OK successful run.
@@ -398,8 +339,6 @@ NJS_EXPORT njs_int_t njs_vm_run(njs_vm_t *vm);
  */
 NJS_EXPORT njs_int_t njs_vm_start(njs_vm_t *vm, njs_value_t *retval);
 
-NJS_EXPORT njs_int_t njs_vm_add_path(njs_vm_t *vm, const njs_str_t *path);
-
 #define NJS_PROTO_ID_ANY    (-1)
 
 NJS_EXPORT njs_int_t njs_vm_external_prototype(njs_vm_t *vm,
@@ -441,10 +380,13 @@ NJS_EXPORT njs_int_t njs_vm_bind(njs_vm_t *vm, const njs_str_t *var_name,
 njs_int_t njs_vm_bind_handler(njs_vm_t *vm, const njs_str_t *var_name,
     njs_prop_handler_t handler, uint16_t magic16, uint32_t magic32,
     njs_bool_t shared);
+NJS_EXPORT njs_int_t njs_vm_global(njs_vm_t *vm, njs_value_t *retval);
 NJS_EXPORT njs_int_t njs_vm_value(njs_vm_t *vm, const njs_str_t *path,
     njs_value_t *retval);
 NJS_EXPORT njs_function_t *njs_vm_function(njs_vm_t *vm, const njs_str_t *name);
 NJS_EXPORT njs_bool_t njs_vm_constructor(njs_vm_t *vm);
+NJS_EXPORT njs_int_t njs_vm_prototype(njs_vm_t *vm, njs_value_t *value,
+    njs_value_t *retval);
 
 NJS_EXPORT void njs_vm_throw(njs_vm_t *vm, const njs_value_t *value);
 NJS_EXPORT void njs_vm_error2(njs_vm_t *vm, unsigned error_type,
@@ -460,20 +402,10 @@ NJS_EXPORT njs_int_t njs_value_to_integer(njs_vm_t *vm, njs_value_t *value,
 
 /*  Gets string value, no copy. */
 NJS_EXPORT void njs_value_string_get(njs_value_t *value, njs_str_t *dst);
-/*
- * Sets a byte string value.
- *   start data is not copied and should not be freed.
- */
-NJS_EXPORT njs_int_t njs_vm_value_string_set(njs_vm_t *vm, njs_value_t *value,
-    const u_char *start, uint32_t size);
-NJS_EXPORT u_char *njs_vm_value_string_alloc(njs_vm_t *vm, njs_value_t *value,
-    uint32_t size);
 NJS_EXPORT njs_int_t njs_vm_value_string_create(njs_vm_t *vm,
     njs_value_t *value, const u_char *start, uint32_t size);
 NJS_EXPORT njs_int_t njs_vm_value_string_create_chb(njs_vm_t *vm,
     njs_value_t *value, njs_chb_t *chain);
-NJS_EXPORT njs_int_t njs_vm_value_string_copy(njs_vm_t *vm, njs_str_t *retval,
-    njs_value_t *value, uintptr_t *next);
 NJS_EXPORT njs_int_t njs_vm_string_compare(const njs_value_t *v1,
     const njs_value_t *v2);
 
@@ -522,9 +454,6 @@ NJS_EXPORT njs_int_t njs_vm_value_dump(njs_vm_t *vm, njs_str_t *dst,
 
 NJS_EXPORT void njs_vm_memory_error(njs_vm_t *vm);
 
-NJS_EXPORT void njs_vm_logger(njs_vm_t *vm, njs_log_level_t level,
-    const char *fmt, ...);
-
 NJS_EXPORT void njs_value_undefined_set(njs_value_t *value);
 NJS_EXPORT void njs_value_null_set(njs_value_t *value);
 NJS_EXPORT void njs_value_invalid_set(njs_value_t *value);
@@ -540,6 +469,7 @@ NJS_EXPORT double njs_value_number(const njs_value_t *value);
 NJS_EXPORT njs_function_t *njs_value_function(const njs_value_t *value);
 NJS_EXPORT njs_function_native_t njs_value_native_function(
     const njs_value_t *value);
+NJS_EXPORT void *njs_value_ptr(const njs_value_t *value);
 njs_external_ptr_t njs_value_external(const njs_value_t *value);
 NJS_EXPORT njs_int_t njs_value_external_tag(const njs_value_t *value);
 
@@ -569,6 +499,10 @@ NJS_EXPORT njs_int_t njs_vm_object_alloc(njs_vm_t *vm, njs_value_t *retval,
     ...);
 NJS_EXPORT njs_value_t *njs_vm_object_keys(njs_vm_t *vm, njs_value_t *value,
     njs_value_t *retval);
+NJS_EXPORT njs_value_t *njs_vm_value_enumerate(njs_vm_t *vm, njs_value_t *value,
+    uint32_t flags, njs_value_t *retval);
+NJS_EXPORT njs_value_t *njs_vm_value_own_enumerate(njs_vm_t *vm,
+    njs_value_t *value, uint32_t flags, njs_value_t *retval);
 NJS_EXPORT njs_value_t *njs_vm_object_prop(njs_vm_t *vm,
     njs_value_t *value, const njs_str_t *key, njs_opaque_value_t *retval);
 NJS_EXPORT njs_int_t njs_vm_object_prop_set(njs_vm_t *vm, njs_value_t *value,
